import os
import pickle

import torch.utils.data as data
from litsr.utils import load_file_list, read_image
from tqdm import tqdm


class ImageFolder(data.Dataset):
    """
    Construct a dataset from a folder
    """

    def __init__(self, datapath, repeat=1, cache=None, first_k=None):
        self.datapath = datapath
        self.repeat = repeat or 1
        self.cache = cache

        if cache == "bin":
            [self.filenames, self.files] = self.load_data_from_bin(first_k)
        elif cache == "memory":
            [self.filenames, self.files] = self.load_data_from_memery(first_k)
        else:
            [self.filenames, self.files] = self.load_data()

        if first_k:
            self.filenames = self.filenames[:first_k]
            self.files = self.files[:first_k]

    def __getitem__(self, idx):
        x = self.files[idx % len(self.files)]

        if self.cache == "memory":
            return x
        elif self.cache == "bin":
            with open(x, "rb") as f:
                x = pickle.load(f)
            return x
        else:
            return read_image(x)

    def load_data_from_bin(self, first_k=None):
        file_names = load_file_list(self.datapath, regx=".*.png")
        file_names = [os.path.basename(n) for n in file_names]

        if first_k:
            file_names = file_names[:first_k]

        bin_root = os.path.join(
            os.path.dirname(self.datapath), "_bin_" + os.path.basename(self.datapath)
        )
        if not os.path.exists(bin_root):
            os.mkdir(bin_root)
            print("mkdir", bin_root)
        files = []

        for f in file_names:
            bin_file = os.path.join(bin_root, f.split(".")[0] + ".pkl")
            if not os.path.exists(bin_file):
                with open(bin_file, "wb") as bin_f:
                    pickle.dump(
                        read_image(os.path.join(self.datapath, f)), bin_f,
                    )
                print("dump", bin_file)
            files.append(bin_file)
        return file_names, files

    def load_data_from_memery(self, first_k=None, from_pickle=True):
        if from_pickle:
            file_names, filenames = self.load_data_from_bin(first_k)
            files = []
            pbar = tqdm(filenames)
            pbar.set_description("load data (from pickle)")
            for f in pbar:
                with open(f, "rb") as f_:
                    files.append(pickle.load(f_))

            return file_names, files

        file_names = load_file_list(self.datapath, regx=".*.png")
        file_names = [os.path.basename(n) for n in file_names]

        if first_k:
            file_names = file_names[:first_k]

        files = []
        pbar = tqdm(file_names)
        pbar.set_description("load data")
        for f in pbar:
            files.append(read_image(os.path.join(self.datapath, f)))

        return file_names, files

    def load_data(self):
        files = load_file_list(self.datapath, regx=".*.png")
        file_names = [os.path.basename(n) for n in files]
        return file_names, files

    def __len__(self):
        return len(self.files) * self.repeat


class PairedImageFolder(data.Dataset):
    def __init__(self, path1, path2, repeat=1, cache=None, first_k=None):
        self.dataset_1 = ImageFolder(path1, repeat=repeat, cache=cache, first_k=first_k)
        self.dataset_2 = ImageFolder(path2, repeat=repeat, cache=cache, first_k=first_k)

        self.filenames = self.dataset_2.filenames

        assert len(self.dataset_1) == len(self.dataset_2)

    def __getitem__(self, i):
        return tuple([self.dataset_1[i], self.dataset_2[i]])

    def __len__(self):
        return len(self.dataset_1)


class ConcatDataset(data.Dataset):
    def __init__(self, datasets):
        self.datasets = datasets

    def __getitem__(self, i):
        return tuple(d[i] for d in self.datasets)

    def __len__(self):
        return min(len(d) for d in self.datasets)



####################
# ema helper
####################


def accumulate(model1, model2, decay=0.9999):
    par1 = dict(model1.named_parameters())
    par2 = dict(model2.named_parameters())

    for k in par1.keys():
        par1[k].data.mul_(decay).add_(par2[k].data, alpha=1 - decay)



import os
import importlib
from os import path as osp
from litsr.utils.registry import ModelRegistry

__all__ = ["create_model", "load_model"]

# Import all models
model_folder = osp.dirname(osp.abspath(__file__))
for root, dirs, files in os.walk(model_folder):
    for f in files:
        if f.endswith("_model.py"):
            importlib.import_module(f"litsr.models.{osp.splitext(f)[0]}")


def create_model(opt):
    """create model from option"""
    model_name = opt.get("which_model")
    if not model_name:
        return
    model = ModelRegistry.get(model_name)
    return model(opt)


def load_model(opt, ckpt_path=None, strict=True):
    """load model from checkpoint path"""
    model_name = opt.get("which_model")
    if not model_name:
        return
    model = ModelRegistry.get(model_name)
    model = model.load_from_checkpoint(ckpt_path, opt=opt, strict=strict)
    print("Model loaded from {0}".format(ckpt_path))
    return model



import torch
import torch.nn as nn
import numpy as np
import time
from torch.utils.data import DataLoader
from torch.nn import functional as F
from torchvision.utils import make_grid
import pytorch_lightning as pl
from litsr.archs import create_net
from litsr.utils.registry import ModelRegistry

# from models import forward_self_ensemble
# from datasets.ms_dataset import TrainDataset, TestDataset
# from utils.metrics import calc_metrics
# from utils.utils import Tensor2np


# @ModelRegistry.register()
class MultiScaleSRModel(pl.LightningModule):
    def __init__(self, opt):
        """
        hparams: in_channels, out_channels, num_features, num_blocks, num_layers
        """
        super().__init__()

        # save hyperparameters
        self.save_hyperparameters(opt)

        # init super-resolution network
        self.sr_net = create_net(opt["model"])

        # define loss function (L1 loss)
        self.loss_fn = nn.L1Loss()

    def setup(self, mode):
        # setup the train dataset
        if mode == "fit":
            self.train_dataset = TrainDataset(
                self.hparams.train.data_path,
                self.hparams.train.min_scale,
                self.hparams.train.max_scale,
                self.hparams.train.lr_img_sz,
                self.hparams.train.batch_size,
                self.hparams.model.rgb_range,
                self.hparams.train.data_repeat,
                self.hparams.train.data_cache,
                self.hparams.train.data_first_k,
            )
        self.test_dataset = lambda scale: TestDataset(
            self.hparams.test.data_path, scale, self.hparams.model.rgb_range
        )

    def forward(self, lr, out_size):
        return self.sr_net(lr, out_size)

    def forward_self_ensemble(self, lr, out_size):
        return forward_self_ensemble(self, lr, out_size)

    def training_step(self, batch, batch_idx):
        lr, hr = batch
        out_size = hr.shape[2]
        sr = self.forward(lr, out_size)
        loss = self.loss_fn(sr, hr)
        return loss

    def training_epoch_end(self, outputs):
        if "ran" in self.hparams.model.which_network:
            self.sr_net.update_temperature()
        avg_loss = torch.stack([out["loss"] for out in outputs]).mean()
        self.logger.experiment.add_scalar("train_loss", avg_loss, self.global_step)

    def validation_step(self, batch, batch_idx, dataloader_idx):
        lr, hr, name = batch
        out_size = hr.shape[2:]
        torch.cuda.synchronize()
        start = time.time()
        if self.hparams.test.self_ensemble:
            sr = self.forward_self_ensemble(lr, out_size).detach()
        else:
            sr = self.forward(lr, out_size).detach()
        torch.cuda.synchronize()
        end = time.time()
        crop_border = int(np.ceil(float(hr.shape[2]) / lr.shape[2]))
        loss = self.loss_fn(sr, hr).item()

        sr_np, hr_np = Tensor2np(
            [sr.cpu()[0], hr.cpu()[0]], self.hparams.model.rgb_range
        )
        psnr, ssim = calc_metrics(sr_np, hr_np, crop_border=crop_border, test_Y=True)
        return {
            "val_loss": loss,
            "val_psnr": psnr,
            "val_ssim": ssim,
            "log_img": sr_np,
            "name": name[0],
            "time": end - start,
        }

    def validation_epoch_end(self, outputs):
        psnr_list = []
        for idx, dataloader_output_result in enumerate(outputs):
            scale = self.hparams.valid.scales[idx]
            dataloader_outs = dataloader_output_result
            tensorboard = self.logger.experiment

            log_imgs = [x["log_img"] for x in dataloader_outs[12:13]]

            avg_loss = np.array([x["val_loss"] for x in dataloader_outs]).mean()
            avg_psnr = np.array([x["val_psnr"] for x in dataloader_outs]).mean()
            psnr_list.append(avg_psnr)
            tensorboard.add_scalar(
                "val_loss/{0}".format(scale), avg_loss, self.global_step
            )
            tensorboard.add_scalar(
                "val_psnr/{0}".format(scale), avg_psnr, self.global_step
            )
            if log_imgs:
                tensorboard.add_image(
                    "images/{0}".format(scale),
                    log_imgs[0],
                    self.global_step,
                    dataformats="HWC",
                )

        self.log(
            "avg_val_psnr",
            np.array(psnr_list).mean(),
            on_epoch=True,
            prog_bar=True,
            logger=False,
        )

    def configure_optimizers(self):
        if hasattr(self.hparams.train, "betas"):
            betas = self.hparams.train.betas
        else:
            betas = (0.9, 0.999)
        optimizer = torch.optim.Adam(
            self.parameters(), lr=self.hparams.train.lr, betas=betas
        )
        if hasattr(self.hparams.train, "lr_scheduler_step"):
            LR_scheduler = torch.optim.lr_scheduler.StepLR(
                optimizer,
                step_size=self.hparams.train.lr_scheduler_step,
                gamma=self.hparams.train.lr_scheduler_gamma,
            )
        elif hasattr(self.hparams.train, "lr_scheduler_milestones"):
            LR_scheduler = torch.optim.lr_scheduler.MultiStepLR(
                optimizer,
                milestones=self.hparams.train.lr_scheduler_milestones,
                gamma=self.hparams.train.lr_scheduler_gamma,
            )
        else:
            raise Exception(
                "No lr settings found in self.hparams.train, please modify the config file"
            )
        return [optimizer], [LR_scheduler]

    def train_dataloader(self):
        self.train_dataset.resample_scale()
        return DataLoader(
            self.train_dataset,
            batch_size=self.hparams.train.batch_size,
            num_workers=self.hparams.sys.num_workers,
            shuffle=False,
            pin_memory=True,
        )

    def val_dataloader(self):
        dataloaders = []
        for scale in self.hparams.valid.scales:
            dataset = self.test_dataset(scale)
            dataloaders.append(
                DataLoader(
                    dataset,
                    batch_size=1,
                    num_workers=self.hparams.sys.num_workers,
                    pin_memory=True,
                )
            )
        return dataloaders

    def change_params(self, hparams):
        self.hparams = hparams



# Modified from https://github.com/xinntao/Real-ESRGAN/blob/HEAD/realesrgan/data/realesrgan_dataset.py

import math
import random

import cv2
import numpy as np
import torch
from litsr.data.image_folder import ImageFolder
from litsr.transforms import single2tensor, uint2single, augment
from litsr.utils.registry import DatasetRegistry
from torch.utils import data

from .degradations import circular_lowpass_kernel, random_mixed_kernels


@DatasetRegistry.register()
class RealESRGANDataset(data.Dataset):
    """
    Dataset used for Real-ESRGAN model.
    """

    def __init__(
        self,
        datapath,
        scale,
        is_train,
        degradation_options,
        lr_img_sz=None,
        rgb_range=1,
        repeat=1,
        cache=None,
        first_k=None,
        mean=None,
        std=None,
        return_img_name=False,
    ):
        self.scale = scale
        self.lr_img_sz = lr_img_sz
        self.repeat = repeat or 1
        self.rgb_range = rgb_range or 1
        assert self.rgb_range == 1
        self.is_train = is_train
        self.mean = mean
        self.std = std
        self.return_img_name = return_img_name
        self.dataset = ImageFolder(
            datapath, repeat=self.repeat, cache=cache, first_k=first_k
        )
        self.file_names = self.dataset.filenames
        self.opt = degradation_options

        # blur settings for the first degradation
        self.blur_kernel_size = self.opt["blur_kernel_size"]
        self.kernel_list = self.opt["kernel_list"]
        self.kernel_prob = self.opt["kernel_prob"]
        self.blur_sigma = self.opt["blur_sigma"]
        self.betag_range = self.opt["betag_range"]
        self.betap_range = self.opt["betap_range"]
        self.sinc_prob = self.opt["sinc_prob"]

        # blur settings for the second degradation
        self.blur_kernel_size2 = self.opt["blur_kernel_size2"]
        self.kernel_list2 = self.opt["kernel_list2"]
        self.kernel_prob2 = self.opt["kernel_prob2"]
        self.blur_sigma2 = self.opt["blur_sigma2"]
        self.betag_range2 = self.opt["betag_range2"]
        self.betap_range2 = self.opt["betap_range2"]
        self.sinc_prob2 = self.opt["sinc_prob2"]

        # a final sinc filter
        self.final_sinc_prob = self.opt["final_sinc_prob"]

        self.kernel_range = [
            2 * v + 1 for v in range(3, 11)
        ]  # kernel size ranges from 7 to 21
        self.pulse_tensor = torch.zeros(
            21, 21
        ).float()  # convolving with pulse tensor brings no blurry effect
        self.pulse_tensor[10, 10] = 1

    def __getitem__(self, idx):
        img_gt = self.dataset[idx]  # Uint8
        img_gt = uint2single(img_gt)

        # -------------------- augmentation for training: flip, rotation -------------------- #
        img_gt = augment(img_gt, self.opt["use_hflip"], self.opt["use_rot"])

        # crop or pad to 400: 400 is hard-coded. You may change it accordingly
        h, w = img_gt.shape[0:2]
        crop_pad_size = 400
        # pad
        if h < crop_pad_size or w < crop_pad_size:
            pad_h = max(0, crop_pad_size - h)
            pad_w = max(0, crop_pad_size - w)
            img_gt = cv2.copyMakeBorder(
                img_gt, 0, pad_h, 0, pad_w, cv2.BORDER_REFLECT_101
            )
        # crop
        if img_gt.shape[0] > crop_pad_size or img_gt.shape[1] > crop_pad_size:
            h, w = img_gt.shape[0:2]
            # randomly choose top and left coordinates
            top = random.randint(0, h - crop_pad_size)
            left = random.randint(0, w - crop_pad_size)
            img_gt = img_gt[top : top + crop_pad_size, left : left + crop_pad_size, ...]

        # Generate kernels (used in the first degradation)
        kernel_size = random.choice(self.kernel_range)
        if np.random.uniform() < self.opt["sinc_prob"]:
            # this sinc filter setting is for kernels ranging from [7, 21]
            if kernel_size < 13:
                omega_c = np.random.uniform(np.pi / 3, np.pi)
            else:
                omega_c = np.random.uniform(np.pi / 5, np.pi)
            kernel = circular_lowpass_kernel(omega_c, kernel_size, pad_to=False)
        else:
            kernel = random_mixed_kernels(
                self.kernel_list,
                self.kernel_prob,
                kernel_size,
                self.blur_sigma,
                self.blur_sigma,
                [-math.pi, math.pi],
                self.betag_range,
                self.betap_range,
                noise_range=None,
            )
        # pad kernel
        pad_size = (21 - kernel_size) // 2
        kernel = np.pad(kernel, ((pad_size, pad_size), (pad_size, pad_size)))

        # Generate kernels (used in the second degradation)
        kernel_size = random.choice(self.kernel_range)
        if np.random.uniform() < self.opt["sinc_prob2"]:
            if kernel_size < 13:
                omega_c = np.random.uniform(np.pi / 3, np.pi)
            else:
                omega_c = np.random.uniform(np.pi / 5, np.pi)
            kernel2 = circular_lowpass_kernel(omega_c, kernel_size, pad_to=False)
        else:
            kernel2 = random_mixed_kernels(
                self.kernel_list2,
                self.kernel_prob2,
                kernel_size,
                self.blur_sigma2,
                self.blur_sigma2,
                [-math.pi, math.pi],
                self.betag_range2,
                self.betap_range2,
                noise_range=None,
            )

        # pad kernel
        pad_size = (21 - kernel_size) // 2
        kernel2 = np.pad(kernel2, ((pad_size, pad_size), (pad_size, pad_size)))

        # sinc kernel
        if np.random.uniform() < self.opt["final_sinc_prob"]:
            kernel_size = random.choice(self.kernel_range)
            omega_c = np.random.uniform(np.pi / 3, np.pi)
            sinc_kernel = circular_lowpass_kernel(omega_c, kernel_size, pad_to=21)
            sinc_kernel = torch.FloatTensor(sinc_kernel)
        else:
            sinc_kernel = self.pulse_tensor

        # BGR to RGB, HWC to CHW, numpy to tensor
        img_gt = single2tensor(img_gt)
        kernel = torch.FloatTensor(kernel)
        kernel2 = torch.FloatTensor(kernel2)

        return_d = {
            "gt": img_gt,
            "kernel1": kernel,
            "kernel2": kernel2,
            "sinc_kernel": sinc_kernel,
        }
        return return_d

    def __len__(self):
        return len(self.dataset)



import json
import os
import re
import shutil
from pathlib import Path
from typing import Iterable, Optional

import cv2
import numpy as np
import yaml
from easydict import EasyDict as Dict
from litsr.transforms import uint2single
from tqdm import tqdm


def read_json(fname):
    """Read json from path
    """
    fname = Path(fname)
    with fname.open("rt") as handle:
        return Dict(json.load(handle))


def write_json(content, fname):
    """Write json to path
    """
    fname = Path(fname)
    with fname.open("wt") as handle:
        json.dump(content, handle, indent=4, sort_keys=False)


def read_yaml(fname):
    """Real yaml from path

    Args:
        fname (str): path

    Returns:
        Easydict: Easydict
    """
    fname = Path(fname)
    with fname.open("rt") as handle:
        return Dict(yaml.load(handle, Loader=yaml.FullLoader))


def mkdir(path):
    """create a single empty directory if it didn't exist

    Parameters:
        path (str) -- a single directory path
    """
    if not os.path.exists(path):
        os.makedirs(path)


def mkdirs(paths):
    """create empty directories if they don't exist

    Parameters:
        paths (str list) -- a list of directory paths
    """
    if isinstance(paths, list) and not isinstance(paths, str):
        for path in paths:
            mkdir(path)
    else:
        mkdir(paths)


def mkdir_clean(path):
    """Create a directory that is guaranteed to be empty 
    """
    if os.path.exists(path):
        try:
            shutil.rmtree(path)
            os.makedirs(path)
        except:
            print("warning! Cannot remove {0}".format(path))
    else:
        os.makedirs(path)


def load_file_list(path: str, regx: str) -> list:
    file_list = os.listdir(path)
    return_list = []
    for _, f in enumerate(file_list):
        if re.search(regx, f):
            return_list.append(os.path.join(path, f))
    return sorted(return_list)


def read_image(
    path: str, mode: Optional[str] = "RGB", to_float: bool = False
) -> np.ndarray:
    """Read image to a 3 dimentional numpy by OpenCV
    """
    img = cv2.imread(path)
    assert mode in ("RGB", "BGR")
    if mode == "RGB":
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

    if to_float:
        img = uint2single(img)
    return img


def read_images(
    path_list: Iterable[str], mode: Optional[str] = "RGB", to_float: bool = False
) -> np.ndarray:
    """Read images to a 4 dimentional numpy array by OpenCV
    """
    rslt = []
    for path in tqdm(path_list):
        rslt.append(read_image(path, mode, to_float))
    return np.array(rslt)



# import all data module

from litsr.utils.registry import DataModuleRegistry, DatasetRegistry

from .downsampled_dataset import DownsampledDataset
from .image_folder import ImageFolder, PairedImageFolder
from .paired_image_dataset import PairedImageDataset
from .realesrgan_dataset import RealESRGANDataset
from .realsr_datamodule import RealSRDataModule
from .sr_datamodule import SRDataModule

__all__ = ["create_data_module"]


def create_data_module(opt):
    """ create model from option
    """
    data_module_name = opt.get("data_module")
    if not data_module_name:
        return
    data_module = DataModuleRegistry.get(data_module_name)
    return data_module(opt)



import importlib
import os
from os import path as osp

import torch
from litsr.utils.registry import ArchRegistry

__all__ = ["create_net", "load_net", "load_or_create_net", "freeze", "unfreeze"]

# Import all archs
arch_folder = osp.dirname(osp.abspath(__file__))
for root, dirs, files in os.walk(arch_folder):
    for f in files:
        if f.endswith("_arch.py"):
            importlib.import_module(f"litsr.archs.{osp.splitext(f)[0]}")


def create_net(opt):
    """Create Network from option
    """
    name = opt["name"]
    args = opt["args"]

    Net = ArchRegistry.get(name)
    net = Net(**args)

    return net


def load_net(opt, state_dict_getter=None):
    """Load network parameters from chekcpoint
    """
    if not opt.get("pretrained_path"):
        raise Exception("pretrained path not given")

    state_dict = torch.load(opt.pretrained_path)

    if not state_dict_getter and hasattr(state_dict, "state_dict"):
        # default state_dict_getter
        state_dict = state_dict["state_dict"]
        # new_state_dict = OrderedDict()
        # for k, v in state_dict.items():
        #     k = k.replace("sr_net.", "")
        #     new_state_dict[k] = v
    elif state_dict_getter:
        state_dict = state_dict_getter(state_dict)

    net = create_net(opt)

    net.load_state_dict(state_dict, strict=True)
    print(
        "Model {0} loaded from {1}".format(net.__class__.__name__, opt.pretrained_path)
    )
    return net


def load_or_create_net(opt, state_dict_getter=None):
    if not opt.get("pretrained_path"):
        return create_net(opt)
    else:
        return load_net(opt, state_dict_getter)


def freeze(module):
    if not isinstance(module, torch.nn.Module):
        raise ValueError("input must be torch.nn.Module")
    for p in module.parameters():
        p.requires_grad = False
    print("Model {0} has been frozen. ".format(module.__class__.__name__))


def unfreeze(module):
    if not isinstance(module, torch.nn.Module):
        raise ValueError("input must be torch.nn.Module")
    for p in module.parameters():
        p.requires_grad = True
    print("Model {0} has been unfrozen. ".format(module.__class__.__name__))



# A translation of https://github.com/hojonathanho/diffusion/blob/master/diffusion_tf/diffusion_utils_2.py
# from TensorFlow with some help from https://github.com/rosinality/denoising-diffusion-pytorch

import numbers
import torch
import numpy as np
from torch import nn
from torch.nn import functional as F


def extract(input, t, shape):
    out = torch.gather(input, 0, t.to(input.device))
    reshape = [shape[0]] + [1] * (len(shape) - 1)
    out = out.reshape(*reshape)

    return out


def get_beta_schedule(schedule, start, end, n_timestep):
    def _warmup_beta(start, end, n_timestep, warmup_frac):

        betas = end * torch.ones(n_timestep, dtype=torch.float64)
        warmup_time = int(n_timestep * warmup_frac)
        betas[:warmup_time] = torch.linspace(
            start, end, warmup_time, dtype=torch.float64
        )

        return betas

    if schedule == "quad":
        betas = (
            torch.linspace(start ** 0.5, end ** 0.5, n_timestep, dtype=torch.float64)
            ** 2
        )
    elif schedule == "linear":
        betas = torch.linspace(start, end, n_timestep, dtype=torch.float64)
    elif schedule == "warmup10":
        betas = _warmup_beta(start, end, n_timestep, 0.1)
    elif schedule == "warmup50":
        betas = _warmup_beta(start, end, n_timestep, 0.5)
    elif schedule == "const":
        betas = end * torch.ones(n_timestep, dtype=torch.float64)
    elif schedule == "jsd":  # 1/T, 1/(T-1), 1/(T-2), ..., 1
        betas = 1.0 / (torch.linspace(n_timestep, 1, n_timestep, dtype=torch.float64))
    else:
        raise NotImplementedError(schedule)

    return betas


class Diffusion(nn.Module):
    def __init__(
        self,
        beta_type,
        beta_start,
        beta_end,
        n_timestep,
        model_var_type,
        model_mean_type,
        **kwargs
    ):
        super().__init__()

        betas = get_beta_schedule(beta_type, beta_start, beta_end, n_timestep).type(
            torch.float64
        )

        timesteps = betas.shape[0]
        self.num_timesteps = int(timesteps)

        self.model_var_type = model_var_type  # learned, fixedsmall, fixedlarge
        self.model_mean_type = model_mean_type  # xprev, xstart, eps

        alphas = 1 - betas
        alphas_cumprod = torch.cumprod(alphas, 0)
        alphas_cumprod_prev = torch.cat(
            (torch.tensor([1], dtype=torch.float64), alphas_cumprod[:-1]), 0
        )
        posterior_variance = betas * (1 - alphas_cumprod_prev) / (1 - alphas_cumprod)

        self.register("betas", betas)
        self.register("alphas_cumprod", alphas_cumprod)
        self.register("alphas_cumprod_prev", alphas_cumprod_prev)

        self.register("sqrt_alphas_cumprod", torch.sqrt(alphas_cumprod))
        self.register("sqrt_one_minus_alphas_cumprod", torch.sqrt(1 - alphas_cumprod))
        self.register("log_one_minus_alphas_cumprod", torch.log(1 - alphas_cumprod))
        self.register("sqrt_recip_alphas_cumprod", torch.rsqrt(alphas_cumprod))
        self.register("sqrt_recipm1_alphas_cumprod", torch.sqrt(1 / alphas_cumprod - 1))
        self.register("posterior_variance", posterior_variance)
        self.register(
            "posterior_log_variance_clipped",
            torch.log(
                torch.cat(
                    (
                        posterior_variance[1].view(1, 1),
                        posterior_variance[1:].view(-1, 1),
                    ),
                    0,
                )
            ).view(-1),
        )
        self.register(
            "posterior_mean_coef1",
            (betas * torch.sqrt(alphas_cumprod_prev) / (1 - alphas_cumprod)),
        )
        self.register(
            "posterior_mean_coef2",
            ((1 - alphas_cumprod_prev) * torch.sqrt(alphas) / (1 - alphas_cumprod)),
        )

    def register(self, name, tensor):
        self.register_buffer(name, tensor.type(torch.float32))

    def q_mean_variance(self, x_0, t):
        mean = extract(self.sqrt_alphas_cumprod, t, x_0.shape) * x_0
        variance = extract(1.0 - self.alphas_cumprod, t, x_0.shape)
        log_variance = extract(self.log_one_minus_alphas_cumprod, t, x_0.shape)
        return mean, variance, log_variance

    def q_sample(self, x_0, t, noise=None):
        if noise is None:
            noise = torch.randn_like(x_0)

        return (
            extract(self.sqrt_alphas_cumprod, t, x_0.shape) * x_0
            + extract(self.sqrt_one_minus_alphas_cumprod, t, x_0.shape) * noise
        )

    def q_posterior_mean_variance(self, x_0, x_t, t):
        mean = (
            extract(self.posterior_mean_coef1, t, x_t.shape) * x_0
            + extract(self.posterior_mean_coef2, t, x_t.shape) * x_t
        )
        var = extract(self.posterior_variance, t, x_t.shape)
        log_var_clipped = extract(self.posterior_log_variance_clipped, t, x_t.shape)

        return mean, var, log_var_clipped

    def p_mean_variance(self, model, x, t, ctx, clip_denoised, return_pred_x0):

        model_output = model(x, t, ctx)

        # Learned or fixed variance?
        if self.model_var_type == "learned":
            model_output, log_var = torch.split(model_output, 2, dim=-1)
            var = torch.exp(log_var)

        elif self.model_var_type in ["fixedsmall", "fixedlarge"]:

            # below: only log_variance is used in the KL computations
            var, log_var = {
                # for 'fixedlarge', we set the initial (log-)variance like so to get a better decoder log likelihood
                "fixedlarge": (
                    self.betas,
                    torch.log(
                        torch.cat(
                            (
                                self.posterior_variance[1].view(1, 1),
                                self.betas[1:].view(-1, 1),
                            ),
                            0,
                        )
                    ).view(-1),
                ),
                "fixedsmall": (
                    self.posterior_variance,
                    self.posterior_log_variance_clipped,
                ),
            }[self.model_var_type]

            var = extract(var, t, x.shape) * torch.ones_like(x)
            log_var = extract(log_var, t, x.shape) * torch.ones_like(x)
        else:
            raise NotImplementedError(self.model_var_type)

        # Mean parameterization
        _maybe_clip = lambda x_: (x_.clamp(min=-1, max=1) if clip_denoised else x_)

        if self.model_mean_type == "xprev":
            # the model predicts x_{t-1}
            pred_x0 = _maybe_clip(
                self.predict_start_from_prev(x_t=x, t=t, x_prev=model_output)
            )
            mean = model_output
        elif self.model_mean_type == "xstart":
            # the model predicts x_0
            pred_x0 = _maybe_clip(model_output)
            mean, _, _ = self.q_posterior_mean_variance(x_0=pred_x0, x_t=x, t=t)
        elif self.model_mean_type == "eps":
            # the model predicts epsilon
            pred_x0 = _maybe_clip(
                self._predict_start_from_noise(x_t=x, t=t, noise=model_output)
            )
            mean, _, _ = self.q_posterior_mean_variance(x_0=pred_x0, x_t=x, t=t)
        else:
            raise NotImplementedError(self.model_mean_type)

        if return_pred_x0:
            return mean, var, log_var, pred_x0
        else:
            return mean, var, log_var

    def _predict_start_from_noise(self, x_t, t, noise):

        return (
            extract(self.sqrt_recip_alphas_cumprod, t, x_t.shape) * x_t
            - extract(self.sqrt_recipm1_alphas_cumprod, t, x_t.shape) * noise
        )

    def _predict_noise_from_start(self, x_t, t, pred_xstart):
        return (
            extract(self.sqrt_recip_alphas_cumprod, t, x_t.shape) * x_t - pred_xstart
        ) / extract(self.sqrt_recipm1_alphas_cumprod, t, x_t.shape)

    def p_sample(
        self, model, x, t, ctx, noise_fn, clip_denoised=True, return_pred_x0=False
    ):

        mean, _, log_var, pred_x0 = self.p_mean_variance(
            model, x, t, ctx, clip_denoised, return_pred_x0=True
        )
        noise = noise_fn(x.shape, dtype=x.dtype).to(x.device)

        shape = [x.shape[0]] + [1] * (x.ndim - 1)
        nonzero_mask = (1 - (t == 0).type(torch.float32)).view(*shape).to(x.device)
        sample = mean + nonzero_mask * torch.exp(0.5 * log_var) * noise

        return (sample, pred_x0) if return_pred_x0 else sample

    @torch.no_grad()
    def p_sample_loop(self, model, ctx, shape, noise_fn=torch.randn):

        device = "cuda" if next(model.parameters()).is_cuda else "cpu"
        img = noise_fn(shape).to(device)

        for i in reversed(range(self.num_timesteps)):
            img = self.p_sample(
                model,
                img,
                torch.full((shape[0],), i, dtype=torch.int64).to(device),
                ctx=ctx,
                noise_fn=noise_fn,
                return_pred_x0=False,
            )

        return img

    @torch.no_grad()
    def p_sample_loop_fast(
        self,
        model,
        ctx,
        shape,
        device,
        noise_fn,
        skip=10,
        eta=0.0,
        include_x0_pred_freq=10,
    ):

        seq = range(0, self.num_timesteps, skip)
        seq_next = [0] + list(seq[:-1])

        x = noise_fn(shape).to(device)

        x0_preds = []
        for idx, i, j in zip(range(len(seq)), reversed(seq), reversed(seq_next)):
            t = torch.full((shape[0],), i, dtype=torch.int64).to(device)
            t_prev = torch.full((shape[0],), j, dtype=torch.int64).to(device)

            alpha_bar = extract(self.alphas_cumprod, t, x.shape)
            alpha_bar_prev = extract(self.alphas_cumprod, t_prev, x.shape)

            _, _, _, pred_x0 = self.p_mean_variance(
                model, x, t, ctx, True, return_pred_x0=True
            )

            if idx % include_x0_pred_freq == 0:
                x0_preds.append(pred_x0)

            eps = self._predict_noise_from_start(x, t, pred_x0)

            sigma = (
                eta
                * torch.sqrt((1 - alpha_bar_prev) / (1 - alpha_bar))
                * torch.sqrt(1 - alpha_bar / alpha_bar_prev)
            )

            noise = torch.randn_like(x)

            mean_pred = (
                pred_x0 * torch.sqrt(alpha_bar_prev)
                + torch.sqrt((1 - alpha_bar_prev) - sigma ** 2) * eps
            )

            nonzero_mask = (
                (t != 0).float().view(-1, *([1] * (len(x.shape) - 1)))
            )  # no noise when t == 0

            x = mean_pred + nonzero_mask * sigma * noise

        return x, torch.cat(x0_preds, dim=0)

    @torch.no_grad()
    def p_sample_loop_progressive(
        self, model, ctx, shape, device, noise_fn=torch.randn, include_x0_pred_freq=50
    ):

        img = noise_fn(shape, dtype=torch.float32).to(device)

        num_recorded_x0_pred = self.num_timesteps // include_x0_pred_freq
        x0_preds_ = torch.zeros(
            (shape[0], num_recorded_x0_pred, *shape[1:]), dtype=torch.float32
        ).to(device)

        for i in reversed(range(self.num_timesteps)):

            # Sample p(x_{t-1} | x_t) as usual
            img, pred_x0 = self.p_sample(
                model=model,
                x=img,
                t=torch.full((shape[0],), i, dtype=torch.int64).to(device),
                ctx=ctx,
                noise_fn=noise_fn,
                return_pred_x0=True,
            )

            # Keep track of prediction of x0
            insert_mask = np.floor(i // include_x0_pred_freq) == torch.arange(
                num_recorded_x0_pred, dtype=torch.int32, device=device
            )

            insert_mask = insert_mask.to(torch.float32).view(
                1, num_recorded_x0_pred, *([1] * len(shape[1:]))
            )
            x0_preds_ = (
                insert_mask * pred_x0[:, None, ...] + (1.0 - insert_mask) * x0_preds_
            )

        return img, x0_preds_

    def get_loss(self, model, x_0, t, ctx, noise=None, return_pred_x0=False):

        if noise is None:
            noise = torch.randn_like(x_0)

        x_t = self.q_sample(x_0=x_0, t=t, noise=noise)

        target = {
            "xprev": self.q_posterior_mean_variance(x_0=x_0, x_t=x_t, t=t)[0],
            "xstart": x_0,
            "eps": noise,
        }[self.model_mean_type]

        model_output = model(x_t, t, ctx)

        if return_pred_x0:
            _maybe_clip = lambda x_: x_.clamp(min=-1, max=1)

            if self.model_mean_type == "xprev":
                pred_x0 = _maybe_clip(
                    self.predict_start_from_prev(x_t=x_t, t=t, x_prev=model_output)
                )
            elif self.model_mean_type == "xstart":
                pred_x0 = _maybe_clip(model_output)
            elif self.model_mean_type == "eps":
                pred_x0 = _maybe_clip(
                    self._predict_start_from_noise(x_t=x_t, t=t, noise=model_output)
                )
            else:
                raise NotImplementedError(self.model_mean_type)

        loss = torch.mean((target - model_output).view(x_0.shape[0], -1) ** 2, dim=1)

        return (loss, pred_x0) if return_pred_x0 else loss



import torch.utils.data as data
from litsr import transforms
from litsr.data.image_folder import PairedImageFolder
from litsr.utils.registry import DatasetRegistry


@DatasetRegistry.register()
class PairedImageDataset(data.Dataset):
    def __init__(
        self,
        lr_path,
        hr_path,
        scale,
        is_train,
        lr_img_sz=None,
        rgb_range=1,
        repeat=1,
        cache=None,
        first_k=None,
        mean=None,
        std=None,
        return_img_name=False,
    ):
        assert not is_train ^ bool(lr_img_sz)

        self.dataset = PairedImageFolder(
            lr_path, hr_path, repeat=repeat, cache=cache, first_k=first_k
        )
        self.scale = scale
        self.lr_img_sz = lr_img_sz
        self.repeat = repeat
        self.rgb_range = rgb_range
        self.is_train = is_train
        self.mean = mean
        self.std = std
        self.return_img_name = return_img_name
        self.file_names = self.dataset.filenames

    def __getitem__(self, idx):
        lr, hr = self.dataset[idx]
        if self.is_train:
            lr, hr = self._transform_train(lr, hr)
        else:
            lr, hr = self._transform_test(lr, hr)

        if self.mean and self.std:
            transforms.normalize(lr, self.mean, self.std, inplace=True)
            transforms.normalize(hr, self.mean, self.std, inplace=True)

        if self.return_img_name:
            file_name = self.file_names[idx % (len(self.dataset) // self.repeat)]
            return lr, hr, file_name
        else:
            return lr, hr

    def _transform_train(self, lr, hr):
        scale = self.scale
        lr_img_sz = self.lr_img_sz
        hr_img_sz = int(lr_img_sz * scale)

        hr = transforms.mod_crop(hr, self.scale)
        lr, hr = transforms.paired_random_crop(lr, hr, hr_img_sz, self.scale)
        lr, hr = transforms.augment([lr, hr])
        lr, hr = (
            transforms.uint2single(lr) * self.rgb_range,
            transforms.uint2single(hr) * self.rgb_range,
        )
        lr, hr = transforms.single2tensor(lr), transforms.single2tensor(hr)

        return lr, hr

    def _transform_test(self, lr, hr):
        scale = self.scale
        hr = transforms.mod_crop(hr, scale)

        lr, hr = (
            transforms.uint2single(lr) * self.rgb_range,
            transforms.uint2single(hr) * self.rgb_range,
        )
        lr, hr = transforms.single2tensor(lr), transforms.single2tensor(hr)

        return lr, hr

    def __len__(self):
        return len(self.dataset)



import numpy as np
import math
from skimage import color as sc
import cv2

####################
# metric
####################
def calc_psnr_ssim(img1, img2, crop_border, test_Y=True):
    #
    # img1 = img1 / 255.
    # img2 = img2 / 255.

    if test_Y and img1.shape[2] == 3:  # evaluate on Y channel in YCbCr color space
        im1_in = sc.rgb2ycbcr(img1)[:, :, 0]
        im2_in = sc.rgb2ycbcr(img2)[:, :, 0]
    else:
        im1_in = img1
        im2_in = img2
    if im1_in.ndim == 3:
        cropped_im1 = im1_in[crop_border:-crop_border, crop_border:-crop_border, :]
        cropped_im2 = im2_in[crop_border:-crop_border, crop_border:-crop_border, :]
    elif im1_in.ndim == 2:
        cropped_im1 = im1_in[crop_border:-crop_border, crop_border:-crop_border]
        cropped_im2 = im2_in[crop_border:-crop_border, crop_border:-crop_border]
    else:
        raise ValueError(
            "Wrong image dimension: {}. Should be 2 or 3.".format(im1_in.ndim)
        )

    psnr = calc_psnr(cropped_im1, cropped_im2)
    ssim = calc_ssim(cropped_im1, cropped_im2)
    return psnr, ssim


def calc_psnr(img1, img2):
    img1 = img1.astype(np.float64)
    img2 = img2.astype(np.float64)
    mse = np.mean((img1 - img2) ** 2)
    if mse == 0:
        return float("inf")
    return 20 * math.log10(255.0 / math.sqrt(mse))


def ssim(img1, img2):

    C1 = (0.01 * 255) ** 2
    C2 = (0.03 * 255) ** 2

    img1 = img1.astype(np.float64)
    img2 = img2.astype(np.float64)
    kernel = cv2.getGaussianKernel(11, 1.5)
    window = np.outer(kernel, kernel.transpose())

    mu1 = cv2.filter2D(img1, -1, window)[5:-5, 5:-5]  # valid
    mu2 = cv2.filter2D(img2, -1, window)[5:-5, 5:-5]
    mu1_sq = mu1 ** 2
    mu2_sq = mu2 ** 2
    mu1_mu2 = mu1 * mu2
    sigma1_sq = cv2.filter2D(img1 ** 2, -1, window)[5:-5, 5:-5] - mu1_sq
    sigma2_sq = cv2.filter2D(img2 ** 2, -1, window)[5:-5, 5:-5] - mu2_sq
    sigma12 = cv2.filter2D(img1 * img2, -1, window)[5:-5, 5:-5] - mu1_mu2

    ssim_map = ((2 * mu1_mu2 + C1) * (2 * sigma12 + C2)) / (
        (mu1_sq + mu2_sq + C1) * (sigma1_sq + sigma2_sq + C2)
    )
    return ssim_map.mean()


def calc_ssim(img1, img2):

    """calculate SSIM
    the same outputs as MATLAB's
    img1, img2: [0, 255]
    """
    if not img1.shape == img2.shape:
        raise ValueError("Input images must have the same dimensions.")
    if img1.ndim == 2:
        return ssim(img1, img2)
    elif img1.ndim == 3:
        if img1.shape[2] == 3:
            ssims = []
            for i in range(3):
                ssims.append(ssim(img1, img2))
            return np.array(ssims).mean()
        elif img1.shape[2] == 1:
            return ssim(np.squeeze(img1), np.squeeze(img2))
    else:
        raise ValueError("Wrong input image dimensions.")



import random
from typing import Iterable, List, Optional, Union

import cv2
import numpy as np
import torch
from PIL import Image
from torch import Tensor

from . import matlab_funcs

#####################
# convert functions #
#####################


def uint2single(img: np.ndarray) -> np.ndarray:
    """Convert uint8 to float32
    """

    assert img.dtype == np.uint8
    return np.float32(img / 255.0)


def single2uint(img: np.ndarray) -> np.ndarray:
    """Convert float32 to uint8
    """

    return np.uint8((img.clip(0, 1) * 255.0).round())


def single2tensor(img: np.ndarray) -> np.ndarray:
    """Convert float32 to tensor
    """

    return torch.from_numpy(np.ascontiguousarray(img)).permute(2, 0, 1).float()


def pil2tensor(
    imgs: Union[Image.Image, Iterable[Image.Image]], rgb_range: Optional[int] = "255"
) -> torch.tensor:
    """Convert a pillow image to tensor
    """

    def _pil2tensor(img, rgb_range):
        img = np.array(img)
        np_transpose = np.ascontiguousarray(img.transpose((2, 0, 1)))
        tensor = torch.from_numpy(np_transpose).float()
        tensor.mul_(rgb_range / 255.0)
        return tensor

    if not isinstance(imgs, list):
        return _pil2tensor(imgs, rgb_range)
    else:
        return [_pil2tensor(img, rgb_range) for img in imgs]


def tensor2uint8(
    tensors: Union[torch.tensor, Iterable[torch.tensor]], rgb_range: Optional[int] = 255
):
    """Convert tensor (0, rgb_range) to numpy array
    """
    to_list_flag = 0
    if not isinstance(tensors, list):
        to_list_flag = 1
        tensors = [tensors]

    def quantize(img, rgb_range):
        pixel_range = 255.0 / rgb_range
        return img.mul(pixel_range).clamp(0, 255).round()

    array = [
        np.transpose(quantize(tensor, rgb_range).numpy(), (1, 2, 0)).astype(np.uint8)
        for tensor in tensors
    ]
    if to_list_flag:
        array = array[0]

    return array


def mod_crop(img, scale):
    """Mod crop images, used during testing.

    Args:
        img (ndarray): Input image.
        scale (int): Scale factor.

    Returns:
        ndarray: Result image.
    """
    img = img.copy()
    if img.ndim in (2, 3):
        h, w = img.shape[0], img.shape[1]
        h_remainder, w_remainder = h % scale, w % scale
        img = img[: h - h_remainder, : w - w_remainder, ...]
    else:
        raise ValueError(f"Wrong img ndim: {img.ndim}.")
    return img


def random_crop(imgs, patch_size):
    if not isinstance(imgs, list):
        imgs = [imgs]

    input_type = "Tensor" if torch.is_tensor(imgs[0]) else "Numpy"

    if input_type == "Tensor":
        h, w = imgs[0].size()[-2:]
    else:
        h, w = imgs[0].shape[0:2]

    if h < patch_size or w < patch_size:
        raise ValueError(
            f"LQ ({h}, {w}) is smaller than patch size "
            f"({patch_size}, {patch_size}). "
        )

    # randomly choose top and left coordinates for patch
    top = random.randint(0, h - patch_size)
    left = random.randint(0, w - patch_size)

    # crop lq patch
    if input_type == "Tensor":
        imgs = [v[:, :, top : top + patch_size, left : left + patch_size] for v in imgs]
    else:
        imgs = [v[top : top + patch_size, left : left + patch_size, ...] for v in imgs]

    if len(imgs) == 1:
        return imgs[0]
    else:
        return imgs


def paired_random_crop(img_lqs, img_gts, gt_patch_size, scale, gt_path=None):
    """Paired random crop. Support Numpy array and Tensor inputs.

    It crops lists of lq and gt images with corresponding locations.

    Args:
        img_gts (list[ndarray] | ndarray | list[Tensor] | Tensor): GT images. Note that all images
            should have the same shape. If the input is an ndarray, it will
            be transformed to a list containing itself.
        img_lqs (list[ndarray] | ndarray): LQ images. Note that all images
            should have the same shape. If the input is an ndarray, it will
            be transformed to a list containing itself.
        gt_patch_size (int): GT patch size.
        scale (int): Scale factor.
        gt_path (str): Path to ground-truth. Default: None.

    Returns:
        list[ndarray] | ndarray: GT images and LQ images. If returned results
            only have one element, just return ndarray.
    """

    if not isinstance(img_gts, list):
        img_gts = [img_gts]
    if not isinstance(img_lqs, list):
        img_lqs = [img_lqs]

    # determine input type: Numpy array or Tensor
    input_type = "Tensor" if torch.is_tensor(img_gts[0]) else "Numpy"

    if input_type == "Tensor":
        h_lq, w_lq = img_lqs[0].size()[-2:]
        h_gt, w_gt = img_gts[0].size()[-2:]
    else:
        h_lq, w_lq = img_lqs[0].shape[0:2]
        h_gt, w_gt = img_gts[0].shape[0:2]
    lq_patch_size = gt_patch_size // scale

    if h_gt != h_lq * scale or w_gt != w_lq * scale:
        raise ValueError(
            f"Scale mismatches. GT ({h_gt}, {w_gt}) is not {scale}x ",
            f"multiplication of LQ ({h_lq}, {w_lq}).",
        )
    if h_lq < lq_patch_size or w_lq < lq_patch_size:
        raise ValueError(
            f"LQ ({h_lq}, {w_lq}) is smaller than patch size "
            f"({lq_patch_size}, {lq_patch_size}). "
            f"Please remove {gt_path}."
        )

    # randomly choose top and left coordinates for lq patch
    top = random.randint(0, h_lq - lq_patch_size)
    left = random.randint(0, w_lq - lq_patch_size)

    # crop lq patch
    if input_type == "Tensor":
        img_lqs = [
            v[:, :, top : top + lq_patch_size, left : left + lq_patch_size]
            for v in img_lqs
        ]
    else:
        img_lqs = [
            v[top : top + lq_patch_size, left : left + lq_patch_size, ...]
            for v in img_lqs
        ]

    # crop corresponding gt patch
    top_gt, left_gt = int(top * scale), int(left * scale)
    if input_type == "Tensor":
        img_gts = [
            v[:, :, top_gt : top_gt + gt_patch_size, left_gt : left_gt + gt_patch_size]
            for v in img_gts
        ]
    else:
        img_gts = [
            v[top_gt : top_gt + gt_patch_size, left_gt : left_gt + gt_patch_size, ...]
            for v in img_gts
        ]
    if len(img_gts) == 1:
        img_gts = img_gts[0]
    if len(img_lqs) == 1:
        img_lqs = img_lqs[0]
    return img_gts, img_lqs


def vflip(img):
    """vertical flip
    """
    return cv2.flip(img, 0, img)


def random_vflip(img, p=0.5):
    """Random vertival flips
    """
    if random.random() < p:
        return vflip(img)
    else:
        return img


def hflip(img):
    """horizontal flip
    """
    return cv2.flip(img, 1, img)


def random_hflip(img, p=0.5):
    """Random horizontal flips
    """
    if random.random() < p:
        return hflip(img)
    else:
        return img


def rot90(img):
    """rotate 90
    """
    return img.transpose(1, 0, 2)


def augment(imgs, hflip=True, rotation=True, return_status=False):
    """Augment: horizontal flips OR rotate (0, 90, 180, 270 degrees).

    We use vertical flip and transpose for rotation implementation.
    All the images in the list use the same augmentation.

    Args:
        imgs (list[ndarray] | ndarray): Images to be augmented. If the input
            is an ndarray, it will be transformed to a list.
        hflip (bool): Horizontal flip. Default: True.
        rotation (bool): Ratotation. Default: True.
        return_status (bool): Return the status of flip and rotation.
            Default: False.

    Returns:
        list[ndarray] | ndarray: Augmented images and flows. If returned
            results only have one element, just return ndarray.

    """
    hflip = hflip and random.random() < 0.5
    vflip = rotation and random.random() < 0.5
    rot90 = rotation and random.random() < 0.5

    def _augment(img):
        img = img.copy()
        if hflip:  # horizontal
            cv2.flip(img, 1, img)
        if vflip:  # vertical
            cv2.flip(img, 0, img)
        if rot90:
            img = img.transpose(1, 0, 2)
        return img

    if not isinstance(imgs, list):
        imgs = [imgs]
    imgs = [_augment(img) for img in imgs]
    if len(imgs) == 1:
        imgs = imgs[0]

    if return_status:
        return imgs, (hflip, vflip, rot90)
    else:
        return imgs


def img_rotate(img, angle, center=None, scale=1.0):
    """Rotate image.

    Args:
        img (ndarray): Image to be rotated.
        angle (float): Rotation angle in degrees. Positive values mean
            counter-clockwise rotation.
        center (tuple[int]): Rotation center. If the center is None,
            initialize it as the center of the image. Default: None.
        scale (float): Isotropic scale factor. Default: 1.0.
    """
    (h, w) = img.shape[:2]

    if center is None:
        center = (w // 2, h // 2)

    matrix = cv2.getRotationMatrix2D(center, angle, scale)
    rotated_img = cv2.warpAffine(img, matrix, (w, h))
    return rotated_img


def resize_cv2(img, scale=None, size=None, mode="bicubic"):
    """Resize image use opencv

    Args:
        img (np.array): Input image.
        scale (float | None): scale factor.
        mode ('bicubic' | 'bilinear' | 'nearest'): interpolate method.

    Return:
        img
    """
    assert scale or size
    h, w = img.shape[0:2]
    if not size:
        size = (int(h * scale), int(w * scale))

    size = (size[1], size[0])

    interpolation = {
        "bicubic": cv2.INTER_CUBIC,
        "bilinear": cv2.INTER_LINEAR,
        "nearest": cv2.INTER_NEAREST,
    }[mode]

    return cv2.resize(img, size, interpolation=interpolation)


def resize_matlab(img, scale=None, size=None, mode="bicubic"):
    """Resize function the same as matlab
    
    Args:
        img (Tensor | np.array): Input image.
        scale (float): scale factor.
        mode ('bicubic'): matlab function only support bicubic interpolation.

    """
    assert mode == "bicubic" and (scale or size)
    to_uint_flag = 0
    if img.dtype == np.uint8:
        to_uint_flag = 1
        img = uint2single(img)
    img = matlab_funcs.imresize(img, scale, size)
    if to_uint_flag:
        img = single2uint(img)
    return img


def resize_pillow(img, scale=None, size=None, mode="bicubic"):
    from PIL import Image

    assert scale or size
    h, w = img.size[0:2]
    if not size:
        size = (int(h * scale), int(w * scale))

    size = (size[1], size[0])
    to_numpy_flag = 0
    if type(img) == np.ndarray:
        to_numpy_flag = 1
        img = Image.fromarray(img)

    interpolation = {
        "bicubic": Image.BICUBIC,
        "bilinear": Image.BILINEAR,
        "nearest": Image.NEAREST,
    }[mode]

    img = img.resize(size, resample=interpolation)
    if to_numpy_flag:
        img = np.array(img)

    return img


def normalize(
    tensor: Tensor, mean: List[float], std: List[float], inplace: bool = False
) -> Tensor:
    """Normalize a tensor image with mean and standard deviation.
    This transform does not support PIL Image.

    .. note::
        This transform acts out of place by default, i.e., it does not mutates the input tensor.

    See :class:`~torchvision.transforms.Normalize` for more details.

    Args:
        tensor (Tensor): Tensor image of size (C, H, W) or (B, C, H, W) to be normalized.
        mean (sequence): Sequence of means for each channel.
        std (sequence): Sequence of standard deviations for each channel.
        inplace(bool,optional): Bool to make this operation inplace.

    Returns:
        Tensor: Normalized Tensor image.
    """
    if not isinstance(tensor, torch.Tensor):
        raise TypeError(
            "Input tensor should be a torch tensor. Got {}.".format(type(tensor))
        )

    if tensor.ndim < 3:
        raise ValueError(
            "Expected tensor to be a tensor image of size (..., C, H, W). Got tensor.size() = "
            "{}.".format(tensor.size())
        )

    if not inplace:
        tensor = tensor.clone()

    dtype = tensor.dtype
    mean = torch.as_tensor(mean, dtype=dtype, device=tensor.device)
    std = torch.as_tensor(std, dtype=dtype, device=tensor.device)
    if (std == 0).any():
        raise ValueError(
            "std evaluated to zero after conversion to {}, leading to division by zero.".format(
                dtype
            )
        )
    if mean.ndim == 1:
        mean = mean.view(-1, 1, 1)
    if std.ndim == 1:
        std = std.view(-1, 1, 1)
    tensor.sub_(mean).div_(std)
    return tensor


def denormalize(
    tensor: Tensor, mean: List[float], std: List[float], inplace: bool = False
) -> Tensor:
    if not isinstance(tensor, torch.Tensor):
        raise TypeError(
            "Input tensor should be a torch tensor. Got {}.".format(type(tensor))
        )

    if tensor.ndim < 3:
        raise ValueError(
            "Expected tensor to be a tensor image of size (..., C, H, W). Got tensor.size() = "
            "{}.".format(tensor.size())
        )

    if not inplace:
        tensor = tensor.clone()

    dtype = tensor.dtype
    mean = torch.as_tensor(mean, dtype=dtype, device=tensor.device)
    std = torch.as_tensor(std, dtype=dtype, device=tensor.device)
    if (std == 0).any():
        raise ValueError(
            "std evaluated to zero after conversion to {}, leading to division by zero.".format(
                dtype
            )
        )
    if mean.ndim == 1:
        mean = mean.view(-1, 1, 1)
    if std.ndim == 1:
        std = std.view(-1, 1, 1)
    tensor.mul_(std).add_(std)
    return tensor



from typing import Any, Optional

from litsr.utils.registry import DataModuleRegistry
from pytorch_lightning import LightningDataModule
from torch.utils.data import DataLoader

from .downsampled_dataset import DownsampledDataset
from .paired_image_dataset import PairedImageDataset


@DataModuleRegistry.register()
class SRDataModule(LightningDataModule):
    def __init__(self, opt):
        super().__init__()
        self.opt = opt

    def setup(self, stage: Optional[str]):
        if stage == "fit":
            self.train_dataset = DownsampledDataset(
                datapath=self.opt.train.data_path,
                scale=self.opt.train.scale,
                is_train=True,
                repeat=self.opt.train.get("data_repeat"),
                lr_img_sz=self.opt.train.lr_img_sz,
                rgb_range=self.opt.train.get("rgb_range"),
                cache=self.opt.train.get("data_cache"),
                first_k=self.opt.train.get("data_first_k"),
                mean=self.opt.train.get("mean"),
                std=self.opt.train.get("std"),
                downsample_mode=self.opt.train.get("downsample_mode"),
                return_img_name=False,
            )

        self.val_dataset = PairedImageDataset(
            lr_path=self.opt.valid.lr_data_path,
            hr_path=self.opt.valid.hr_data_path,
            scale=self.opt.valid.scale,
            is_train=False,
            rgb_range=self.opt.train.get("rgb_range"),
            mean=self.opt.train.get("mean"),
            std=self.opt.train.get("std"),
            return_img_name=True,
        )

    def train_dataloader(self):
        return DataLoader(
            self.train_dataset,
            batch_size=self.opt.train.batch_size,
            num_workers=self.opt.sys.num_workers,
            shuffle=True,
            pin_memory=True,
        )

    def val_dataloader(self):
        return DataLoader(
            self.val_dataset,
            batch_size=1,
            num_workers=self.opt.sys.num_workers,
            pin_memory=True,
        )

    def test_dataloader(self):
        return DataLoader(
            self.val_dataset,
            batch_size=1,
            num_workers=self.opt.sys.num_workers,
            pin_memory=True,
        )




# Modified from https://github1s.com/xinntao/BasicSR/blob/master/basicsr/data/transforms.py

from .data_funcs import *

__all__ = [
    "mod_crop",
    "random_crop",
    "paired_random_crop",
    "augment",
    "resize_matlab",
    "resize_cv2",
]



import math

import torch
import torch.nn as nn
from torch.nn.utils import weight_norm

################
# Basic blocks
################


def default_conv(in_channels, out_channels, kernel_size, bias=True):
    return nn.Conv2d(
        in_channels, out_channels, kernel_size, padding=(kernel_size // 2), bias=bias
    )


class WeightNormedConv(nn.Sequential):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        bias=True,
        act=nn.ReLU(True),
        res_scale=1.0,
    ):
        conv = weight_norm(
            nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size,
                padding=kernel_size // 2,
                stride=stride,
                bias=bias,
            )
        )
        # init.constant_(conv.weight_g, res_scale)
        # init.zeros_(conv.bias)
        m = [conv]
        if act:
            m.append(act)
        super().__init__(*m)


class MeanShift(nn.Conv2d):
    def __init__(self, rgb_range, rgb_mean, rgb_std, sign=-1):
        super(MeanShift, self).__init__(3, 3, kernel_size=1)
        std = torch.Tensor(rgb_std)
        self.weight.data = torch.eye(3).view(3, 3, 1, 1)
        self.weight.data.div_(std.view(3, 1, 1, 1))
        self.bias.data = sign * rgb_range * torch.Tensor(rgb_mean)
        self.bias.data.div_(std)
        self.requires_grad = False


class BasicBlock(nn.Sequential):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        bias=False,
        bn=True,
        act=nn.ReLU(True),
    ):

        m = [
            nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size,
                padding=(kernel_size // 2),
                stride=stride,
                bias=bias,
            )
        ]
        if bn:
            m.append(nn.BatchNorm2d(out_channels))
        if act is not None:
            m.append(act)
        super(BasicBlock, self).__init__(*m)


class ResBlock(nn.Module):
    def __init__(
        self,
        conv,
        n_feats,
        kernel_size,
        bias=True,
        bn=False,
        act=nn.ReLU(True),
        res_scale=1,
    ):

        super(ResBlock, self).__init__()
        m = []
        for i in range(2):
            m.append(conv(n_feats, n_feats, kernel_size, bias=bias))
            if bn:
                m.append(nn.BatchNorm2d(n_feats))
            if i == 0:
                m.append(act)

        self.body = nn.Sequential(*m)
        self.res_scale = res_scale

    def forward(self, x):
        res = self.body(x).mul(self.res_scale)
        res += x
        return res


def channel_shuffle(x, groups):
    batchsize, num_channels, height, width = x.data.size()

    channels_per_group = num_channels // groups

    # reshape
    x = x.view(batchsize, groups, channels_per_group, height, width)

    x = torch.transpose(x, 1, 2).contiguous()

    # flatten
    x = x.view(batchsize, -1, height, width)

    return x


def make_coord(shape, ranges=None, flatten=True):
    """ Make coordinates at grid centers.
    """
    coord_seqs = []
    for i, n in enumerate(shape):
        if ranges is None:
            v0, v1 = -1, 1
        else:
            v0, v1 = ranges[i]
        r = (v1 - v0) / (2 * n)
        seq = v0 + r + (2 * r) * torch.arange(n).float()
        coord_seqs.append(seq)
    ret = torch.stack(torch.meshgrid(*coord_seqs), dim=-1)
    if flatten:
        ret = ret.view(-1, ret.shape[-1])
    return ret


class TimeEmbedding(nn.Module):
    def __init__(self, dim):
        super().__init__()

        self.dim = dim
        half_dim = self.dim // 2
        self.inv_freq = torch.exp(
            torch.arange(half_dim, dtype=torch.float32)
            * (-math.log(10000) / (half_dim - 1))
        )

    def forward(self, input):
        shape = input.shape
        input = input.view(-1).to(torch.float32)
        sinusoid_in = torch.ger(input, self.inv_freq.to(input.device))
        pos_emb = torch.cat([sinusoid_in.sin(), sinusoid_in.cos()], dim=-1)
        pos_emb = pos_emb.view(*shape, self.dim)

        return pos_emb


@torch.no_grad()
def variance_scaling_init_(tensor, scale=1, mode="fan_avg", distribution="uniform"):
    fan_in, fan_out = nn.init._calculate_fan_in_and_fan_out(tensor)

    if mode == "fan_in":
        scale /= fan_in

    elif mode == "fan_out":
        scale /= fan_out

    else:
        scale /= (fan_in + fan_out) / 2

    if distribution == "normal":
        std = math.sqrt(scale)

        return tensor.normal_(0, std)

    else:
        bound = math.sqrt(3 * scale)

        return tensor.uniform_(-bound, bound)



# Residual Dense Network for Image Super-Resolution
# https://arxiv.org/abs/1802.08797

from . import common

import torch
import torch.nn as nn
from litsr.utils.registry import ArchRegistry


class RDB_Conv(nn.Module):
    def __init__(self, inChannels, growRate, kSize=3):
        super(RDB_Conv, self).__init__()
        Cin = inChannels
        G = growRate
        self.conv = nn.Sequential(
            *[nn.Conv2d(Cin, G, kSize, padding=(kSize - 1) // 2, stride=1), nn.ReLU()]
        )

    def forward(self, x):
        out = self.conv(x)
        return torch.cat((x, out), 1)


class RDB(nn.Module):
    def __init__(self, growRate0, growRate, nConvLayers, kSize=3):
        super(RDB, self).__init__()
        G0 = growRate0
        G = growRate
        C = nConvLayers

        convs = []
        for c in range(C):
            convs.append(RDB_Conv(G0 + c * G, G))
        self.convs = nn.Sequential(*convs)

        # Local Feature Fusion
        self.LFF = nn.Conv2d(G0 + C * G, G0, 1, padding=0, stride=1)

    def forward(self, x):
        return self.LFF(self.convs(x)) + x


@ArchRegistry.register()
class RDN(nn.Module):
    def __init__(
        self,
        scale,
        num_features,
        num_blocks,
        num_layers,
        rgb_range,
        in_channels,
        out_channels,
    ):
        super().__init__()
        r = scale
        G0 = num_features
        kSize = 3

        # number of RDB blocks, conv layers, out channels
        self.D, C, G = [num_blocks, num_layers, num_features]

        rgb_mean = (0.4488, 0.4371, 0.4040)
        rgb_std = (1.0, 1.0, 1.0)
        self.sub_mean = common.MeanShift(rgb_range, rgb_mean, rgb_std)
        self.add_mean = common.MeanShift(rgb_range, rgb_mean, rgb_std, 1)

        # Shallow feature extraction net
        self.SFENet1 = nn.Conv2d(
            in_channels, G0, kSize, padding=(kSize - 1) // 2, stride=1
        )
        self.SFENet2 = nn.Conv2d(G0, G0, kSize, padding=(kSize - 1) // 2, stride=1)

        # Redidual dense blocks and dense feature fusion
        self.RDBs = nn.ModuleList()
        for i in range(self.D):
            self.RDBs.append(RDB(growRate0=G0, growRate=G, nConvLayers=C))

        # Global Feature Fusion
        self.GFF = nn.Sequential(
            *[
                nn.Conv2d(self.D * G0, G0, 1, padding=0, stride=1),
                nn.Conv2d(G0, G0, kSize, padding=(kSize - 1) // 2, stride=1),
            ]
        )

        # Up-sampling net
        if r == 2 or r == 3:
            self.UPNet = nn.Sequential(
                *[
                    nn.Conv2d(G0, G * r * r, kSize, padding=(kSize - 1) // 2, stride=1),
                    nn.PixelShuffle(r),
                    nn.Conv2d(
                        G, out_channels, kSize, padding=(kSize - 1) // 2, stride=1
                    ),
                ]
            )
        elif r == 4:
            self.UPNet = nn.Sequential(
                *[
                    nn.Conv2d(G0, G * 4, kSize, padding=(kSize - 1) // 2, stride=1),
                    nn.PixelShuffle(2),
                    nn.Conv2d(G, G * 4, kSize, padding=(kSize - 1) // 2, stride=1),
                    nn.PixelShuffle(2),
                    nn.Conv2d(
                        G, out_channels, kSize, padding=(kSize - 1) // 2, stride=1
                    ),
                ]
            )

    def forward(self, x):
        x = self.sub_mean(x)
        f__1 = self.SFENet1(x)
        x = self.SFENet2(f__1)

        RDBs_out = []
        for i in range(self.D):
            x = self.RDBs[i](x)
            RDBs_out.append(x)

        x = self.GFF(torch.cat(RDBs_out, 1))
        x += f__1

        x = self.UPNet(x)
        x = self.add_mean(x)
        return x



import time

import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn as nn
from litsr.archs import create_net
from litsr.transforms import tensor2uint8
from litsr.metrics import calc_psnr_ssim
from litsr.utils.registry import ModelRegistry


@ModelRegistry.register()
class SRModel(pl.LightningModule):
    """
    Basic SR Model optimized by pixel-wise loss
    """

    def __init__(self, opt):
        """
        opt: in_channels, out_channels, num_features, num_blocks, num_layers
        """
        super().__init__()

        # save hyperparameters
        self.save_hyperparameters(opt)

        # init super-resolution network
        self.sr_net = create_net(opt["network"])

        # define loss function (L1 loss)
        self.loss_fn = nn.L1Loss()

    def forward(self, lr):
        return self.sr_net(lr)

    def training_step(self, batch, batch_idx):
        lr, hr = batch
        sr = self.forward(lr)
        loss = self.loss_fn(sr, hr)
        return loss

    def training_epoch_end(self, outputs):
        avg_loss = torch.stack([out["loss"] for out in outputs]).mean()
        self.logger.experiment.add_scalar("train_loss", avg_loss, self.global_step)
        return

    def validation_step(self, batch, batch_idx):
        lr, hr, name = batch
        torch.cuda.synchronize()
        start = time.time()
        sr = self.forward(lr).detach()
        torch.cuda.synchronize()
        end = time.time()

        loss = self.loss_fn(sr, hr)
        sr_np, hr_np = tensor2uint8(
            [sr.cpu()[0], hr.cpu()[0]], self.hparams.train.rgb_range
        )

        psnr, ssim = calc_psnr_ssim(
            sr_np, hr_np, crop_border=self.hparams.valid.scale, test_Y=True
        )
        return {
            "val_loss": loss,
            "val_psnr": psnr,
            "val_ssim": ssim,
            "log_img_sr": sr_np,
            "name": name[0],
            "time": end - start,
        }

    def validation_epoch_end(self, outputs):
        tensorboard = self.logger.experiment
        log_img = outputs[0]["log_img_sr"]

        avg_loss = torch.stack([x["val_loss"] for x in outputs]).mean()
        avg_psnr = np.array([x["val_psnr"] for x in outputs]).mean()
        avg_ssim = np.array([x["val_ssim"] for x in outputs]).mean()
        tensorboard.add_scalar("val_loss", avg_loss, self.global_step)
        tensorboard.add_scalar(
            "val_psnr/{0}".format(self.hparams.train.scale), avg_psnr, self.global_step
        )
        tensorboard.add_scalar(
            "val_ssim/{0}".format(self.hparams.train.scale), avg_ssim, self.global_step
        )
        tensorboard.add_image(
            "images/{0}".format(self.hparams.train.scale),
            log_img,
            self.global_step,
            dataformats="HWC",
        )

        self.log("val_psnr", avg_psnr, on_epoch=True, prog_bar=True, logger=False)
        return

    def configure_optimizers(self):
        betas = self.hparams.train.get("betas") or (0.9, 0.999)
        optimizer = torch.optim.Adam(
            self.parameters(), lr=self.hparams.train.lr, betas=betas
        )
        if self.hparams.train.get("lr_scheduler_step"):
            LR_scheduler = torch.optim.lr_scheduler.StepLR(
                optimizer,
                step_size=self.hparams.train.lr_scheduler_step,
                gamma=self.hparams.train.lr_scheduler_gamma,
            )
        elif self.hparams.train.get("lr_scheduler_milestones"):
            LR_scheduler = torch.optim.lr_scheduler.MultiStepLR(
                optimizer,
                milestones=self.hparams.train.lr_scheduler_milestones,
                gamma=self.hparams.train.lr_scheduler_gamma,
            )
        else:
            raise Exception(
                "No lr settings found in self.hparams.train, please modify the config file"
            )
        return [optimizer], [LR_scheduler]



import torch.nn as nn
import math
from litsr.utils.registry import ArchRegistry


def default_conv(in_channels, out_channels, kernel_size, bias=True):
    return nn.Conv2d(
        in_channels, out_channels, kernel_size, padding=(kernel_size // 2), bias=bias
    )


class BasicBlock(nn.Sequential):
    def __init__(
        self,
        conv,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        bias=False,
        bn=True,
        act=nn.ReLU(True),
    ):

        m = [conv(in_channels, out_channels, kernel_size, bias=bias)]
        if bn:
            m.append(nn.BatchNorm2d(out_channels))
        if act is not None:
            m.append(act)

        super(BasicBlock, self).__init__(*m)


class Upsampler(nn.Sequential):
    def __init__(self, conv, scale, n_feats, bn=False, act=False, bias=True):

        m = []
        if (scale & (scale - 1)) == 0:  # Is scale = 2^n?
            for _ in range(int(math.log(scale, 2))):
                m.append(conv(n_feats, 4 * n_feats, 3, bias))
                m.append(nn.PixelShuffle(2))
                if bn:
                    m.append(nn.BatchNorm2d(n_feats))
                if act == "relu":
                    m.append(nn.ReLU(True))
                elif act == "prelu":
                    m.append(nn.PReLU(n_feats))

        elif scale == 3:
            m.append(conv(n_feats, 9 * n_feats, 3, bias))
            m.append(nn.PixelShuffle(3))
            if bn:
                m.append(nn.BatchNorm2d(n_feats))
            if act == "relu":
                m.append(nn.ReLU(True))
            elif act == "prelu":
                m.append(nn.PReLU(n_feats))
        else:
            raise NotImplementedError

        super(Upsampler, self).__init__(*m)


class ResBlock(nn.Module):
    def __init__(
        self,
        conv,
        n_feats,
        kernel_size,
        bias=True,
        bn=False,
        act=nn.ReLU(True),
        res_scale=1,
    ):

        super(ResBlock, self).__init__()
        m = []
        for i in range(2):
            m.append(conv(n_feats, n_feats, kernel_size, bias=bias))
            if bn:
                m.append(nn.BatchNorm2d(n_feats))
            if i == 0:
                m.append(act)

        self.body = nn.Sequential(*m)
        self.res_scale = res_scale

    def forward(self, x):
        res = self.body(x).mul(self.res_scale)
        res += x
        return res


@ArchRegistry.register()
class EDSR(nn.Module):
    def __init__(
        self, n_resblocks, n_feats, scale, in_channels, res_scale=1.0, **kwargs
    ):
        super(EDSR, self).__init__()
        n_colors = 3
        conv = default_conv

        kernel_size = 3
        act = nn.ReLU(True)

        # define head module
        m_head = [conv(in_channels, n_feats, kernel_size)]

        # define body module
        m_body = [
            ResBlock(conv, n_feats, kernel_size, act=act, res_scale=res_scale)
            for _ in range(n_resblocks)
        ]
        m_body.append(conv(n_feats, n_feats, kernel_size))

        self.head = nn.Sequential(*m_head)
        self.body = nn.Sequential(*m_body)

        m_tail = [
            Upsampler(conv, scale, n_feats, act=False),
            conv(n_feats, n_colors, kernel_size),
        ]

        self.tail = nn.Sequential(*m_tail)

    def forward(self, x, output_features=False):
        x = self.head(x)

        res = self.body(x)
        feat = x + res
        out = self.tail(feat)

        if output_features:
            return out, feat
        return out



import torch.utils.data as data
from litsr.data import transforms, utils
from litsr.data.image_folder import ImageFolder
from litsr.utils.registry import DatasetRegistry


@DatasetRegistry.register()
class SingleImageDataset(data.Dataset):
    def __init__(
        self,
        lr_path,
        rgb_range=1,
        repeat=1,
        cache=None,
        first_k=None,
        mean=None,
        std=None,
        return_img_name=False,
    ):

        self.dataset = ImageFolder(lr_path, repeat=repeat, cache=cache, first_k=first_k)
        self.repeat = repeat
        self.rgb_range = rgb_range
        self.mean = mean
        self.std = std
        self.return_img_name = return_img_name
        self.file_names = self.dataset.filenames

    def __getitem__(self, idx):
        lr = self.dataset[idx]

        lr = utils.uint2single(lr)
        lr = utils.single2tensor(lr) * self.rgb_range

        if self.mean and self.std:
            transforms.normalize(lr, self.mean, self.std, inplace=True)

        if self.return_img_name:
            file_name = self.file_names[idx % self.repeat]
            return lr, file_name
        else:
            return lr

    def __len__(self):
        return len(self.paths)



#!/usr/bin/env python

from setuptools import find_packages, setup

import os
import subprocess
import time

version_file = "litsr/version.py"


def readme():
    with open("README.md", encoding="utf-8") as f:
        content = f.read()
    return content


def get_git_hash():
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ["SYSTEMROOT", "PATH", "HOME"]:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env["LANGUAGE"] = "C"
        env["LANG"] = "C"
        env["LC_ALL"] = "C"
        out = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(["git", "rev-parse", "HEAD"])
        sha = out.strip().decode("ascii")
    except OSError:
        sha = "unknown"

    return sha


def get_hash():
    if os.path.exists(".git"):
        sha = get_git_hash()[:7]
    elif os.path.exists(version_file):
        try:
            from litsr.version import __version__

            sha = __version__.split("+")[-1]
        except ImportError:
            raise ImportError("Unable to get git version")
    else:
        sha = "unknown"

    return sha


def write_version_py():
    content = """# GENERATED VERSION FILE
# TIME: {}
__version__ = '{}'
__gitsha__ = '{}'
version_info = ({})
"""
    sha = get_hash()
    with open("VERSION", "r") as f:
        SHORT_VERSION = f.read().strip()
    VERSION_INFO = ", ".join(
        [x if x.isdigit() else f'"{x}"' for x in SHORT_VERSION.split(".")]
    )

    version_file_str = content.format(time.asctime(), SHORT_VERSION, sha, VERSION_INFO)
    with open(version_file, "w") as f:
        f.write(version_file_str)


def get_version():
    with open(version_file, "r") as f:
        exec(compile(f.read(), version_file, "exec"))
    return locals()["__version__"]


def get_requirements(filename="requirements.txt"):
    here = os.path.dirname(os.path.realpath(__file__))
    with open(os.path.join(here, filename), "r") as f:
        requires = [line.replace("\n", "") for line in f.readlines()]
    return requires


if __name__ == "__main__":
    write_version_py()
    setup(
        name="litsr",
        version=get_version(),
        description="Super-Resolution Toolbox Based on PyTorch Lightning",
        long_description=readme(),
        long_description_content_type="text/markdown",
        author="Hanlin Wu",
        author_email="hlwu.bnu@gmail.com",
        keywords="computer vision, super resolution",
        url="",
        include_package_data=True,
        packages=find_packages(exclude=()),
        classifiers=[
            "Development Status :: 3 - Alpha",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
        ],
        license="Apache License 2.0",
        setup_requires=["cython", "numpy"],
        install_requires=get_requirements(),
        zip_safe=False,
    )




import random
import time

import numpy as np
import pytorch_lightning as pl
import torch
from litsr.archs import create_net
from litsr.data.degradations import (
    random_add_gaussian_noise_pt,
    random_add_poisson_noise_pt,
)
from litsr.metrics import calc_psnr_ssim
from litsr.transforms import paired_random_crop, tensor2uint8
from litsr.utils import DiffJPEG, USMSharp, filter2D
from litsr.utils.registry import ModelRegistry
from torch.nn import functional as F


@ModelRegistry.register()
class RealESRNetModel(pl.LightningModule):
    """
    SR Model for real esrgan
    """

    def __init__(self, opt):
        """
        opt: in_channels, out_channels, num_features, num_blocks, num_layers
        """
        super().__init__()

        # save hyperparameters
        self.save_hyperparameters(opt)

        # init super-resolution network
        self.sr_net = create_net(opt["network"])
        self.usm_sharpener = USMSharp()
        self.jpeger = DiffJPEG(differentiable=False)
        self.queue_size = opt.get("queue_size", 180)

        # define loss function (L1 loss)
        self.loss_fn = torch.nn.L1Loss()

    @torch.no_grad()
    def _dequeue_and_enqueue(self):
        # training pair pool
        # initialize
        b, c, h, w = self.lq.size()
        if not hasattr(self, "queue_lr"):
            assert (
                self.queue_size % b == 0
            ), "queue size should be divisible by batch size"
            self.queue_lr = torch.zeros(self.queue_size, c, h, w).cuda()
            _, c, h, w = self.gt.size()
            self.queue_gt = torch.zeros(self.queue_size, c, h, w).cuda()
            self.queue_ptr = 0
        if self.queue_ptr == self.queue_size:  # full
            # do dequeue and enqueue
            # shuffle
            idx = torch.randperm(self.queue_size)
            self.queue_lr = self.queue_lr[idx]
            self.queue_gt = self.queue_gt[idx]
            # get
            lq_dequeue = self.queue_lr[0:b, :, :, :].clone()
            gt_dequeue = self.queue_gt[0:b, :, :, :].clone()
            # update
            self.queue_lr[0:b, :, :, :] = self.lq.clone()
            self.queue_gt[0:b, :, :, :] = self.gt.clone()

            self.lq = lq_dequeue
            self.gt = gt_dequeue
        else:
            # only do enqueue
            self.queue_lr[
                self.queue_ptr : self.queue_ptr + b, :, :, :
            ] = self.lq.clone()
            self.queue_gt[
                self.queue_ptr : self.queue_ptr + b, :, :, :
            ] = self.gt.clone()
            self.queue_ptr = self.queue_ptr + b

    @torch.no_grad()
    def feed_data(self, data, is_train=True):
        if is_train and self.hparams.get("high_order_degradation", True):
            # training data synthesis
            self.gt = data["gt"]
            # USM the GT images
            if self.hparams.get("gt_usm", True):
                self.gt = self.usm_sharpener(self.gt)

            self.kernel1 = data["kernel1"]
            self.kernel2 = data["kernel2"]
            self.sinc_kernel = data["sinc_kernel"]

            ori_h, ori_w = self.gt.size()[2:4]

            # ----------------------- The first degradation process ----------------------- #
            # blur
            out = filter2D(self.gt, self.kernel1)
            # random resize
            updown_type = random.choices(
                ["up", "down", "keep"], self.hparams["resize_prob"]
            )[0]
            if updown_type == "up":
                scale = np.random.uniform(1, self.hparams["resize_range"][1])
            elif updown_type == "down":
                scale = np.random.uniform(self.hparams["resize_range"][0], 1)
            else:
                scale = 1
            mode = random.choice(["area", "bilinear", "bicubic"])
            out = F.interpolate(out, scale_factor=scale, mode=mode)
            # noise
            gray_noise_prob = self.hparams["gray_noise_prob"]
            if np.random.uniform() < self.hparams["gaussian_noise_prob"]:
                out = random_add_gaussian_noise_pt(
                    out,
                    sigma_range=self.hparams["noise_range"],
                    clip=True,
                    rounds=False,
                    gray_prob=gray_noise_prob,
                )
            else:
                out = random_add_poisson_noise_pt(
                    out,
                    scale_range=self.hparams["poisson_scale_range"],
                    gray_prob=gray_noise_prob,
                    clip=True,
                    rounds=False,
                )
            # JPEG compression
            jpeg_p = out.new_zeros(out.size(0)).uniform_(*self.hparams["jpeg_range"])
            out = torch.clamp(out, 0, 1)
            out = self.jpeger(out, quality=jpeg_p)

            # ----------------------- The second degradation process ----------------------- #
            # blur
            if np.random.uniform() < self.hparams["second_blur_prob"]:
                out = filter2D(out, self.kernel2)
            # random resize
            updown_type = random.choices(
                ["up", "down", "keep"], self.hparams["resize_prob2"]
            )[0]
            if updown_type == "up":
                scale = np.random.uniform(1, self.hparams["resize_range2"][1])
            elif updown_type == "down":
                scale = np.random.uniform(self.hparams["resize_range2"][0], 1)
            else:
                scale = 1
            mode = random.choice(["area", "bilinear", "bicubic"])
            out = F.interpolate(
                out,
                size=(
                    int(ori_h / self.hparams["scale"] * scale),
                    int(ori_w / self.hparams["scale"] * scale),
                ),
                mode=mode,
            )
            # noise
            gray_noise_prob = self.hparams["gray_noise_prob2"]
            if np.random.uniform() < self.hparams["gaussian_noise_prob2"]:
                out = random_add_gaussian_noise_pt(
                    out,
                    sigma_range=self.hparams["noise_range2"],
                    clip=True,
                    rounds=False,
                    gray_prob=gray_noise_prob,
                )
            else:
                out = random_add_poisson_noise_pt(
                    out,
                    scale_range=self.hparams["poisson_scale_range2"],
                    gray_prob=gray_noise_prob,
                    clip=True,
                    rounds=False,
                )

            # JPEG compression + the final sinc filter
            # We also need to resize images to desired sizes. We group [resize back + sinc filter] together
            # as one operation.
            # We consider two orders:
            #   1. [resize back + sinc filter] + JPEG compression
            #   2. JPEG compression + [resize back + sinc filter]
            # Empirically, we find other combinations (sinc + JPEG + Resize) will introduce twisted lines.
            if np.random.uniform() < 0.5:
                # resize back + the final sinc filter
                mode = random.choice(["area", "bilinear", "bicubic"])
                out = F.interpolate(
                    out,
                    size=(
                        ori_h // self.hparams["scale"],
                        ori_w // self.hparams["scale"],
                    ),
                    mode=mode,
                )
                out = filter2D(out, self.sinc_kernel)
                # JPEG compression
                jpeg_p = out.new_zeros(out.size(0)).uniform_(
                    *self.hparams["jpeg_range2"]
                )
                out = torch.clamp(out, 0, 1)
                out = self.jpeger(out, quality=jpeg_p)
            else:
                # JPEG compression
                jpeg_p = out.new_zeros(out.size(0)).uniform_(
                    *self.hparams["jpeg_range2"]
                )
                out = torch.clamp(out, 0, 1)
                out = self.jpeger(out, quality=jpeg_p)
                # resize back + the final sinc filter
                mode = random.choice(["area", "bilinear", "bicubic"])
                out = F.interpolate(
                    out,
                    size=(
                        ori_h // self.hparams["scale"],
                        ori_w // self.hparams["scale"],
                    ),
                    mode=mode,
                )
                out = filter2D(out, self.sinc_kernel)

            # clamp and round
            self.lq = torch.clamp((out * 255.0).round(), 0, 255) / 255.0

            # random crop
            gt_size = self.hparams["gt_size"]
            self.gt, self.lq = paired_random_crop(
                self.lq, self.gt, gt_size, self.hparams["scale"]
            )

            # training pair pool
            self._dequeue_and_enqueue()
        else:
            self.lq = data["lq"].to(self.device)
            if "gt" in data:
                self.gt = data["gt"].to(self.device)
                self.gt_usm = self.usm_sharpener(self.gt)

    def forward(self, lr):
        return self.sr_net(lr)

    def training_step(self, batch, batch_idx):
        self.feed_data(batch)
        sr = self.forward(self.lq)
        loss = self.loss_fn(sr, self.gt)
        return loss

    def training_epoch_end(self, outputs):
        avg_loss = torch.stack([out["loss"] for out in outputs]).mean()
        self.logger.experiment.add_scalar("train_loss", avg_loss, self.global_step)
        return

    def validation_step(self, batch, batch_idx):
        lr, hr, name = batch
        torch.cuda.synchronize()
        start = time.time()
        sr = self.forward(lr).detach()
        torch.cuda.synchronize()
        end = time.time()

        loss = self.loss_fn(sr, hr)
        sr_np, hr_np = tensor2uint8(
            [sr.cpu()[0], hr.cpu()[0]], self.hparams.train.rgb_range
        )

        psnr, ssim = calc_psnr_ssim(
            sr_np, hr_np, crop_border=self.hparams.valid.scale, test_Y=True
        )
        return {
            "val_loss": loss,
            "val_psnr": psnr,
            "val_ssim": ssim,
            "log_img_sr": sr_np,
            "name": name[0],
            "time": end - start,
        }

    def validation_epoch_end(self, outputs):
        tensorboard = self.logger.experiment
        log_img = outputs[0]["log_img_sr"]

        avg_loss = torch.stack([x["val_loss"] for x in outputs]).mean()
        avg_psnr = np.array([x["val_psnr"] for x in outputs]).mean()
        avg_ssim = np.array([x["val_ssim"] for x in outputs]).mean()
        tensorboard.add_scalar("val_loss", avg_loss, self.global_step)
        tensorboard.add_scalar(
            "val_psnr/{0}".format(self.hparams.train.scale), avg_psnr, self.global_step
        )
        tensorboard.add_scalar(
            "val_ssim/{0}".format(self.hparams.train.scale), avg_ssim, self.global_step
        )
        tensorboard.add_image(
            "images/{0}".format(self.hparams.train.scale),
            log_img,
            self.global_step,
            dataformats="HWC",
        )

        self.log("val_psnr", avg_psnr, on_epoch=True, prog_bar=True, logger=False)
        return

    def test_step(self, batch, batch_idx):
        return self.validation_step(batch, batch_idx)

    def configure_optimizers(self):
        betas = self.hparams.train.get("betas") or (0.9, 0.999)
        optimizer = torch.optim.Adam(
            self.parameters(), lr=self.hparams.train.lr, betas=betas
        )
        if self.hparams.train.get("lr_scheduler_step"):
            LR_scheduler = torch.optim.lr_scheduler.StepLR(
                optimizer,
                step_size=self.hparams.train.lr_scheduler_step,
                gamma=self.hparams.train.lr_scheduler_gamma,
            )
        elif self.hparams.train.get("lr_scheduler_milestones"):
            LR_scheduler = torch.optim.lr_scheduler.MultiStepLR(
                optimizer,
                milestones=self.hparams.train.lr_scheduler_milestones,
                gamma=self.hparams.train.lr_scheduler_gamma,
            )
        else:
            raise Exception(
                "No lr settings found in self.hparams.train, please modify the config file"
            )
        return [optimizer], [LR_scheduler]



import numpy as np
import torch.utils.data as data
from litsr import transforms
from litsr.data.image_folder import ImageFolder
from litsr.utils.registry import DatasetRegistry
from torchvision import transforms as tv_transforms
from torchvision.transforms import functional as TF


@DatasetRegistry.register()
class DownsampledDataset(data.Dataset):
    def __init__(
        self,
        datapath,
        scale,
        is_train,
        lr_img_sz=None,
        rgb_range=1,
        repeat=1,
        cache=None,
        first_k=None,
        mean=None,
        std=None,
        downsample_mode="bicubic",
        return_img_name=False,
    ):
        assert not is_train ^ bool(lr_img_sz)

        self.scale = scale
        self.lr_img_sz = lr_img_sz
        self.repeat = repeat or 1
        self.rgb_range = rgb_range or 1
        self.is_train = is_train
        self.mean = mean
        self.std = std
        self.return_img_name = return_img_name
        self.downsample_mode = downsample_mode or "bicubic"
        self.dataset = ImageFolder(
            datapath, repeat=self.repeat, cache=cache, first_k=first_k
        )
        self.file_names = self.dataset.filenames

    def __getitem__(self, idx):
        hr = self.dataset[idx]
        if self.is_train:
            lr, hr = self._transform_train(hr)
        else:
            lr, hr = self._transform_test(hr)

        if self.mean and self.std:
            transforms.normalize(lr, self.mean, self.std, inplace=True)
            transforms.normalize(hr, self.mean, self.std, inplace=True)

        if self.return_img_name:
            file_name = self.file_names[idx % (len(self.dataset) // self.repeat)]
            return lr, hr, file_name
        else:
            return lr, hr

    def _transform_train(self, x):
        lr_img_sz = self.lr_img_sz
        hr_img_sz = int(lr_img_sz * self.scale)

        transform = tv_transforms.Compose(
            [
                tv_transforms.ToPILImage(),
                tv_transforms.RandomCrop(hr_img_sz),
                tv_transforms.RandomHorizontalFlip(),
                tv_transforms.RandomVerticalFlip(),
                lambda x: TF.rotate(x, 90) if np.random.random() > 0.5 else x,
                lambda x: [
                    transforms.resize_pillow(
                        x, size=(lr_img_sz, lr_img_sz), mode=self.downsample_mode
                    ),
                    x,
                ],
                lambda x: transforms.pil2tensor(x, self.rgb_range),
            ]
        )
        lr, hr = transform(x)

        return lr, hr

    def _transform_test(self, hr):
        scale = self.scale
        hr = transforms.mod_crop(hr, scale)

        [hr_img_h, hr_img_w] = hr.shape[0:2]
        [lr_img_h, lr_img_w] = [int(hr_img_h // scale), int(hr_img_w // scale)]

        transform = tv_transforms.Compose(
            [
                tv_transforms.ToPILImage(),
                lambda x: [
                    transforms.resize_pillow(
                        x, size=(lr_img_h, lr_img_w), mode=self.downsample_mode
                    ),
                    x,
                ],
                lambda x: transforms.pil2tensor(x, self.rgb_range),
            ]
        )
        lr, hr = transform(hr)

        return lr, hr

    def __len__(self):
        return len(self.dataset)
